
#include "/lib/taa/res_params.glsl"
#include "/lib/uniforms.glsl"

varying vec2 texcoord;

uniform sampler2D colortex13;
uniform sampler2D colortex15;
uniform sampler2D depthtex0;

#include "/lib/color_transforms.glsl"
#include "/lib/projections.glsl"

vec3 tonemap(vec3 col)
{
    return col / (1 + luma(col));
}
vec3 invTonemap(vec3 col)
{
    return col / (1 - luma(col));
}
//  https://software.intel.com/en-us/node/503873
vec3 RGB_YCoCg(vec3 c)
{
    // Y = R/4 + G/2 + B/4
    // Co = R/2 - B/2
    // Cg = -R/4 + G/2 - B/4

    return vec3(c.x / 4.0 + c.y / 2.0 + c.z / 4.0, c.x / 2.0 - c.z / 2.0, -c.x / 4.0 + c.y / 2.0 - c.z / 4.0);
}

// https://software.intel.com/en-us/node/503873
vec3 YCoCg_RGB(vec3 c)
{
    // R = Y + Co - Cg
    // G = Y + Cg
    // B = Y - Co - Cg

    return cout(vec3(c.x + c.y - c.z, c.x + c.z, c.x - c.y - c.z));
}

vec4 textureYC(sampler2D tex, vec2 uv)
{
    vec4 c = texture2D(tex, uv);
    return vec4(RGB_YCoCg(c.rgb), c.a);
}
vec4 texelFetchYC(sampler2D tex, ivec2 uv, float numb)
{
    vec4 c = texelFetch(tex, ivec2(uv), 0);
    return vec4(RGB_YCoCg(c.rgb), c.a);
}

float MinDepth3x3(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(depthtex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(depthtex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}
float MinDepth3x3DH(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(dhDepthTex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(dhDepthTex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}

vec3 ReprojectDH(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }

    return position;
}
vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }

    return position;
}

vec3 clip_aabb(vec3 min_color, vec3 max_color, vec3 history_color)
{
    vec3 p_clip = 0.5 * (max_color + min_color);
    vec3 e_clip = 0.5 * (max_color - min_color);

    vec3 v_clip = history_color - p_clip;
    vec3 v_unit = v_clip / e_clip;
    vec3 a_unit = abs(v_unit);
    float ma_unit = max(a_unit.x, max(a_unit.y, a_unit.z));
    bool history_clipped = ma_unit > 1.0;
    vec3 history = history_clipped ? p_clip + v_clip / ma_unit : history_color;

    return history;
}

void main()
{

    /* RENDERTARGETS: 15 */

    vec3 color = vec3(0.0);
    vec2 uv = texcoord;
    float z = texture(depthtex0, texcoord).x;

    ivec2 texel = ivec2(uv * viewResolution);
    float offcenter_rej = 0.1;
    vec3 position = vec3(uv, MinDepth3x3(texel));
    vec3 reprojectedPosition = Reproject(position);

    vec3 current = textureYC(colortex13, texelSize * texel).rgb;

    vec3 history = textureYC(colortex15, reprojectedPosition.xy).rgb;
    float historyWeight = 0.99;

    if (any(lessThan(reprojectedPosition.xy, vec2(0.0))) || any(greaterThan(reprojectedPosition.xy, vec2(1.0))))
    {
        historyWeight = 0.0;
    }
    vec2 pixelCenterDist = 1.0 - abs(2.0 * fract(reprojectedPosition.xy * viewResolution) - 1.0);
    historyWeight *= sqrt(pixelCenterDist.x * pixelCenterDist.y) * offcenter_rej + (1.0 - offcenter_rej);

    // Read the texel values

    ivec2 du = ivec2(1.0, 0.0);
    ivec2 dv = ivec2(0.0, 1.0);

    vec3 ctl = texelFetchYC(colortex13, (texel - dv - du), 0).rgb;
    vec3 ctc = texelFetchYC(colortex13, (texel - dv), 0).rgb;
    vec3 ctr = texelFetchYC(colortex13, (texel - dv + du), 0).rgb;
    vec3 cml = texelFetchYC(colortex13, (texel - du), 0).rgb;
    vec3 cmc = texelFetchYC(colortex13, (texel), 0).rgb;
    vec3 cmr = texelFetchYC(colortex13, (texel + du), 0).rgb;
    vec3 cbl = texelFetchYC(colortex13, (texel + dv - du), 0).rgb;
    vec3 cbc = texelFetchYC(colortex13, (texel + dv), 0).rgb;
    vec3 cbr = texelFetchYC(colortex13, (texel + dv + du), 0).rgb;
    vec3 minRounded = min(ctl, min(ctc, min(ctr, min(cml, min(cmc, min(cmr, min(cbl, min(cbc, cbr))))))));
    vec3 maxRounded = max(ctl, max(ctc, max(ctr, max(cml, max(cmc, max(cmr, max(cbl, max(cbc, cbr))))))));

    history = clip_aabb(minRounded, maxRounded, history).rgb;

    color = YCoCg_RGB((mix(current, history, historyWeight)));
    color = clamp(color, 0.0001, 1000.0);

    gl_FragData[0].rgb = color;
}
